@@ -4,6 +4,8 @@ import os
 from io import BytesIO
 import sys
 import socket
+import locale
+import re
 import argparse
 import netrc
 import getpass
@@ -11,6 +13,8 @@ import logging
 import time
 import ssl
 
+from dateutil import tz
+
 import paho.mqtt.client
 from PIL import Image
 import ascii_magic
@@ -84,6 +88,9 @@ def main():  # noqa: C901  # pylint: disable=too-many-branches,too-many-statemen
     parser.add_argument('--republish-on-update', dest='republishOnUpdate', action='store_true',
                         help='Republish all topics on every update, not just when the value changes.')
     parser.add_argument('--list-topics', dest='listTopics', help='List new topics when created the first time', action='store_true')
+    parser.add_argument('--topic-filter-regex', dest='topicFilterRegexString', type=str,
+                        default='<PREFIX>/vehicles/[0-9A-Z]+/domains/[a-zA-Z]+/[a-zA-Z]+/request/.*',
+                        help='Filter topics by regex. Default is: "<PREFIX>/vehicles/[0-9A-Z]+/domains/[a-zA-Z]+/[a-zA-Z]+/request/.*"')
 
     weConnectGroup = parser.add_argument_group('WeConnect')
     weConnectGroup.add_argument('-u', '--username', type=str, help='Username of Volkswagen id', required=False)
@@ -103,9 +110,19 @@ def main():  # noqa: C901  # pylint: disable=too-many-branches,too-many-statemen
     weConnectGroup.add_argument('--no-capabilities', dest='noCapabilities', help='Do not add capabilities', action='store_true')
     weConnectGroup.add_argument('--selective', help='Just fetch status of a certain type', default=None, required=False, action='append',
                                 type=domain.Domain, choices=list(domain.Domain))
+    weConnectGroup.add_argument('--convert-times', dest='convertTimes',
+                                help='Convert all times from UTC to timezone, e.g. --convert-times \'Europe/Berlin\', leave empty to use system timezone',
+                                nargs='?', const='', default=None, type=str)
+    weConnectGroup.add_argument('--timeformat', dest='timeFormat',
+                                help='Convert times using the timeformat provided default is ISO format, leave argument empty to use system default',
+                                nargs='?', const='', default=None, type=str)
+    weConnectGroup.add_argument('--locale',
+                                help='Use specified locale, leave argument empty to use system default', default='', type=str)
     parser.add_argument('--pictures', help='Add ASCII art pictures', action='store_true')
     parser.add_argument('--picture-format', dest='pictureFormat', help='Format of the picture topics', default=PictureFormat.TXT, required=False,
                         type=PictureFormat, choices=list(PictureFormat))
+    parser.add_argument('--with-raw-json-topic', dest='withRawJsonTopic', help='Adds topic <PREFIX>/rawjson with all information in one json string.'
+                        ' Topic is updated on change only', action='store_true')
 
     loggingGroup = parser.add_argument_group('Logging')
     loggingGroup.add_argument('-v', '--verbose', action="append_const", help='Logging level (verbosity)', const=-1,)
@@ -116,6 +133,13 @@ def main():  # noqa: C901  # pylint: disable=too-many-branches,too-many-statemen
     loggingGroup.add_argument('--hide-repeated-log', dest='hideRepeatedLog', help='Hide repeated log messages from the same module', action='store_true')
 
     args = parser.parse_args()
+    try:
+        locale.setlocale(locale.LC_ALL, args.locale)
+        if args.timeFormat == '':
+            args.timeFormat = locale.nl_langinfo(locale.D_T_FMT)
+    except locale.Error as err:
+        LOG.error('Cannot set locale: %s', err)
+        sys.exit(1)
 
     logLevel = LOG_LEVELS.index(DEFAULT_LOG_LEVEL)
     for adjustment in args.verbose or ():
@@ -193,10 +217,26 @@ def main():  # noqa: C901  # pylint: disable=too-many-branches,too-many-statemen
                           netRcFilename)
                 sys.exit(1)
 
+    try:
+        topicFilterRegexString = args.topicFilterRegexString
+        topicFilterRegexString.replace('<PREFIX>', args.prefix)
+        topicFilterRegex = re.compile(args.topicFilterRegexString)
+    except re.error as err:
+        LOG.error('Problem with provided regex %s: %s', topicFilterRegexString, err)
+        sys.exit(1)
+
+    convertTimezone = None
+    if args.convertTimes is not None:
+        if args.convertTimes == '':
+            convertTimezone = datetime.now().astimezone().tzinfo
+        else:
+            convertTimezone = tz.gettz(args.convertTimes)
+
     mqttCLient = WeConnectMQTTClient(clientId=args.mqttclientid, transport=args.transport, interval=args.interval,
                                      prefix=args.prefix, ignore=args.ignore, updateCapabilities=(not args.noCapabilities),
                                      updatePictures=args.pictures, selective=args.selective, listNewTopics=args.listTopics,
-                                     republishOnUpdate=args.republishOnUpdate, pictureFormat=args.pictureFormat)
+                                     republishOnUpdate=args.republishOnUpdate, pictureFormat=args.pictureFormat, topicFilterRegex=topicFilterRegex,
+                                     convertTimezone=convertTimezone, timeFormat=args.timeFormat, withRawJsonTopic=args.withRawJsonTopic)
     mqttCLient.enable_logger()
 
     if usetls:
@@ -275,7 +315,7 @@ def main():  # noqa: C901  # pylint: disable=too-many-branches,too-many-statemen
 class WeConnectMQTTClient(paho.mqtt.client.Client):  # pylint: disable=too-many-instance-attributes
     def __init__(self, clientId=None, transport='tcp', interval=300, prefix='weconnect/0', ignore=0,  # pylint: disable=too-many-arguments
                  updateCapabilities=True, updatePictures=True, selective=None, listNewTopics=False, republishOnUpdate=False,
-                 pictureFormat=None):
+                 pictureFormat=None, topicFilterRegex=None, convertTimezone=None, timeFormat=None, withRawJsonTopic=False):
         super().__init__(client_id=clientId, transport=transport)
         self.weConnect = None
         self.prefix = prefix
@@ -294,6 +334,11 @@ class WeConnectMQTTClient(paho.mqtt.client.Client):  # pylint: disable=too-many-
         self.writeableTopics = []
         self.writeableTopicsChanged = False
         self.republishOnUpdate = republishOnUpdate
+        self.topicFilterRegex = topicFilterRegex
+        self.convertTimezone = convertTimezone
+        self.timeFormat = timeFormat
+        self.hasChanges = False
+        self.withRawJsonTopic = withRawJsonTopic
 
         self.on_connect = self.on_connect_callback
         self.on_message = self.on_message_callback
@@ -357,15 +402,23 @@ class WeConnectMQTTClient(paho.mqtt.client.Client):  # pylint: disable=too-many-
         self.setConnected(connected=True)
         self.setError(code=WeConnectErrors.SUCCESS)
 
-    def updateWeConnect(self):
+    def updateWeConnect(self):  # noqa: C901
         LOG.info('Update data from WeConnect')
+        self.hasChanges = False
         try:
             self.weConnect.update(updateCapabilities=self.updateCapabilities, updatePictures=self.updatePictures, selective=self.selective)
             self.setConnected(connected=True)
             self.setError(code=WeConnectErrors.SUCCESS)
             topic = f'{self.prefix}/mqtt/weconnectUpdated'
+            convertedTime = datetime.utcnow().replace(microsecond=0, tzinfo=timezone.utc)
+            if self.convertTimezone is not None:
+                convertedTime = convertedTime.astimezone(self.convertTimezone)
+            if self.timeFormat is not None:
+                convertedTimeString = convertedTime.strftime(self.timeFormat)
+            else:
+                convertedTimeString = convertedTime.isoformat()
             self.publish(topic=topic, qos=1, retain=True,
-                         payload=datetime.utcnow().replace(microsecond=0, tzinfo=timezone.utc).isoformat())
+                         payload=convertedTimeString)
             if topic not in self.topics:
                 self.addTopic(topic)
         except errors.RetrievalError:
@@ -378,16 +431,31 @@ class WeConnectMQTTClient(paho.mqtt.client.Client):  # pylint: disable=too-many-
             errorMessage = f'API compatibility error during update. Will try again after configured interval of {self.interval}s'
             self.setError(code=WeConnectErrors.API_COMPATIBILITY, message=errorMessage)
             LOG.info(errorMessage)
+        except errors.TemporaryAuthentificationError:
+            self.setConnected(connected=False)
+            errorMessage = f'Temporary authentification error during update. Will try again after configured interval of {self.interval}s'
+            self.setError(code=WeConnectErrors.AUTHENTIFICATION, message=errorMessage)
+            LOG.info(errorMessage)
         except socket.error:
             self.setConnected(connected=False)
             errorMessage = f'Socket error during update. Will try again after configured interval of {self.interval}s'
             self.setError(code=WeConnectErrors.RETRIEVAL_FAILED, message=errorMessage)
             LOG.info(errorMessage)
+        if self.withRawJsonTopic and self.hasChanges:
+            topic = f'{self.prefix}/rawjson'
+            if topic not in self.topics:
+                self.addTopic(topic)
+            json = self.weConnect.toJSON()
+            self.publish(topic=topic, qos=1, retain=True, payload=json)
         self.publishTopics()
 
     def onWeConnectEvent(self, element, flags):  # noqa: C901
+        self.hasChanges = True
+        topic = f'{self.prefix}{element.getGlobalAddress()}'
+        if self.topicFilterRegex is not None and self.topicFilterRegex.match(topic):
+            return
+
         if flags & addressable.AddressableLeaf.ObserverEvent.ENABLED:
-            topic = f'{self.prefix}{element.getGlobalAddress()}'
             if isinstance(element, addressable.ChangeableAttribute):
                 topic = topic + '_writetopic'
                 LOG.debug('Subscribe for attribute %s%s', self.prefix, element.getGlobalAddress())
@@ -413,6 +481,13 @@ class WeConnectMQTTClient(paho.mqtt.client.Client):  # pylint: disable=too-many-
             return ', '.join([str(item.value) if isinstance(item, Enum) else str(item) for item in value])
         if isinstance(value, Enum):
             return value.value
+        if isinstance(value, datetime):
+            convertedTime = value
+            if self.convertTimezone is not None:
+                convertedTime = value.astimezone(self.convertTimezone)
+            if self.timeFormat is not None:
+                return convertedTime.strftime(self.timeFormat)
+            return str(convertedTime)
         if isinstance(value, Image.Image):
             if self.pictureFormat == PictureFormat.TXT or self.pictureFormat is None:
                 return util.imgToASCIIArt(value, columns=120, mode=ascii_magic.Modes.ASCII)
